//
// surface.cc
//
//  version 0.4  january 17, 1999
//  jeremy wohl (jeremyw@evantide.com)
//  no warranty, no responsibility, etc.
//

// References:
//
// [Smith95]  : _Image Compositing Fundamentals_, Technical Memo 4, Alvy Ray
//              Smith, August 15, 1995.
// [Foley96]  : _Computer Graphics: Principles and Practice_, 2nd ed.,
//              James D. Foley, et al., 1990.
// [Stevens89]: _Graphics Programming in C_, Roger T. Stevens, 1989.
// [??]       : Bezier.java on the net somewhere.  wrote it?  drop me a note.
//

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "tess.hh"
#include "config.hh"


TessSurface::TessSurface (int width, int height)
{
    init (width, height);
}


TessSurface::TessSurface (TessReader * reader)
{
    reader->get_size (surface_width, surface_height);
    init (surface_width, surface_height);

    reader->populate (this);
}


void TessSurface::init (int width, int height)
{
    surface_width  = width;
    surface_height = height;

    pixmap = new Pixel [surface_width * surface_height];
    memset (pixmap, 0, surface_width * surface_height * sizeof (Pixel));

    type           = 0;
    type_alignment = Left;

    draw_style = Blend;
    draw_width = 0;

    bezier_intervals = 1000;
}


TessSurface::~TessSurface ()
{
    if (pixmap)  delete pixmap;
}


// [Smith95] compositing macros: tess_posit (), etc. in tess.hh
void TessSurface::set_pixel (int x, int y, Pixel& pixel)
{
    Pixel& under = pixmap [y * surface_width + x];
    float  alpha = (float) pixel.alpha / 255.0;


    if (x < 0 || y < 0 || x >= surface_width || y >= surface_height)  return;

    pixel.red    = (ccomp) ((float) pixel.red   * alpha);
    pixel.green  = (ccomp) ((float) pixel.green * alpha);
    pixel.blue   = (ccomp) ((float) pixel.blue  * alpha);

    if (draw_style == Replace || pixel.alpha == 255 || under.alpha == 0)
	under = pixel;
    else
	under = tess_posit (under, pixel);
}


void TessSurface::set_alpha_pixel (int x, int y, ccomp alpha)
{
    Pixel p;

    tess_setp (p, draw_color.red, draw_color.green, draw_color.blue, alpha);

    set_pixel (x, y, p);
}


void TessSurface::get_pixel (int x, int y, Pixel& pixel)
{
    assert (x >= 0 && y >= 0 && x < surface_width && y < surface_height);

    pixel = pixmap [y * surface_width + x];

    float alpha = (float) pixel.alpha / 255.0;
    pixel.red   = (ccomp) ((float) pixel.red   / alpha);
    pixel.green = (ccomp) ((float) pixel.green / alpha);
    pixel.blue  = (ccomp) ((float) pixel.blue  / alpha);
}


void TessSurface::get_pma_pixel (int x, int y, Pixel& pixel)
{
    assert (x >= 0 && y >= 0 && x < surface_width && y < surface_height);

    pixel = pixmap [y * surface_width + x];
}


void TessSurface::set_draw_color (ccomp red, ccomp green, ccomp blue)
{
    tess_setp (draw_color, red, green, blue, 0);
}


void TessSurface::set_draw_width (int width)
{
    draw_width = width;
}


void TessSurface::set_draw_style (TessDrawStyle style)
{
    draw_style = style;
}


void TessSurface::set_type_alignment (TessAlignment alignment)
{
    type_alignment = alignment;
}


void TessSurface::set_type_inclination (int angle)
{
    type_inclination = angle;
}


void TessSurface::draw_string (int x, int y, char * string)
{
    int uc;


    if (! type)  return;

    switch (type_alignment) {
    case Left:
	break;
    case Center:
	x -= type->get_string_width (string) / 2;  break;
    case Right:
	x -= type->get_string_width (string);      break;
    }

    while ((uc = *string++)) {
	type->render_glyph (this, x, y, uc);

	x += type->get_glyph_width (uc) + type->get_spacing ();
    }
}


void TessSurface::draw_uc_string (int x, int y, short * string)
{
    int uc;


    if (! type)  return;

    switch (type_alignment) {
    case Left:
	break;
    case Center:
	x -= type->get_uc_string_width (string) / 2;  break;
    case Right:
	x -= type->get_uc_string_width (string);      break;
    }

    while ((uc = *string++)) {
	type->render_glyph (this, x, y, uc);

	x += type->get_glyph_width (uc) + type->get_spacing ();
    }
}


void TessSurface::draw_line (int x1, int y1, int x2, int y2)
{
    // @@ anti-aliasing support
    draw_naa_line (x1, y1, x2, y2);
}


// [Stevens89]
void TessSurface::draw_naa_line (int x1, int y1, int x2, int y2)
{
    int  dx, dy, dxabs, dyabs, i, px, py, sdx, sdy, x, y;

#define abs(a)  (a < 0 ? -a : a)
#define sign(x) ((x) > 0 ? 1 : ((x) == 0 ? 0 : (-1)))

    dx    =   x2 - x1;  dy    =   y2 - y1;
    sdx   = sign (dx);  sdy   = sign (dy);
    dxabs =  abs (dx);  dyabs =  abs (dy);
    x     =         0;  y     =         0;
    px    =        x1;  py    =        y1;

    if (dxabs > dyabs) {
	for (i = 0; i <= dxabs; i++) {
	    y += dyabs;
	    if (y >= dxabs) {
		y -= dxabs;
		py += sdy;
	    }

	    set_alpha_pixel (px, py, 255);
	    px += sdx;
	}
    }
    else {
	for (i = 0; i <= dyabs; i++) {
	    x += dxabs;
	    if (x >= dyabs) {
		x -= dyabs;
		px += sdx;
	    }

	    set_alpha_pixel (px, py, 255);
	    py += sdy;
	}
    }
}


// [??]
void TessSurface::draw_bezier (int x1, int y1, int x2, int y2, int x3, int y3,
			       int x4, int y4)
{
    double  x, y, xold, yold, A, B, C;
    double  t1, t2, t3;
    double  dx1, dx2, dx3;
    double  dy1, dy2, dy3;
    int     i;


    t1 = 1.0 / bezier_intervals;
    t2 = t1 * t1;
    t3 = t2 * t1;

    x = x1;  xold = x;
    y = y1;  yold = y;

    A =     -x1  + 3 * x2  - 3 * x3  + x4;
    B =  3 * x1  - 6 * x2  + 3 * x3;
    C = -3 * x1  + 3 * x2;

    dx1 =     A * t3  +     B * t2  + C * t1;
    dx2 = 6 * A * t3  + 2 * B * t2;
    dx3 = 6 * A * t3;

    A =     -y1  + 3 * y2  - 3 * y3  + y4;
    B =  3 * y1  - 6 * y2  + 3 * y3;
    C = -3 * y1  + 3 * y2;

    dy1 =     A * t3  +     B * t2  + C * t1;
    dy2 = 6 * A * t3  + 2 * B * t2;
    dy3 = 6 * A * t3;

    for (i = 0; i < bezier_intervals; i++) {
	x += dx1;  dx1 += dx2;	dx2 += dx3;
	y += dy1;  dy1 += dy2;	dy2 += dy3;

	draw_line ((int) xold, (int) yold, (int) x, (int) y);

	xold = x;  yold = y;
    }
}


void TessSurface::draw_rect (int x, int y, int width, int height)
{
    int col, row, x2 = x + width, y2 = y + height;

    
    // top
    for (col = x, row = y;  col < x2; col++)  set_alpha_pixel (col, row, 255);

    // bottom
    for (col = x, row = y2 - 1; col < x2; col++)
	set_alpha_pixel (col, row, 255);

    // left side
    for (col = x, row = y;  row < y2; row++)  set_alpha_pixel (col, row, 255);

    // right side
    for (col = x2 - 1, row = y; row < y2; row++)
	set_alpha_pixel (col, row, 255);
}


void TessSurface::draw_filled_rect (int x, int y, int width, int height)
{
    int col, row, x2 = x + width, y2 = y + height;


    for (row = y; row < y2; row++)
	for (col = x; col < x2; col++)
	    set_alpha_pixel (col, row, 255);
}


void TessSurface::copy_rect (int x, int y, int width, int height,
			     int dest_x, int dest_y)
{
    int    src_col, src_row, x2 = x + width, y2 = y + height;
    int    dest_col, dest_row;
    Pixel  pixel;


    src_row = y;  dest_row = dest_y;

    for ( ; src_row < y2; src_row++, dest_row++) {
	src_col  = x;  dest_col = dest_x;

	for ( ; src_col < x2; src_col++, dest_col++) {
	    get_pixel (src_col,  src_row,  pixel);
	    set_pixel (dest_col, dest_row, pixel);
	}
    }
}


void TessSurface::copy_rect (TessSurface * surface, int x, int y,
			     int width, int height, int dest_x, int dest_y)
{
    int    src_col, src_row, x2 = x + width, y2 = y + height;
    int    dest_col, dest_row;
    Pixel  pixel;


    src_row = y;  dest_row = dest_y;

    for ( ; src_row < y2; src_row++, dest_row++) {
	src_col  = x;  dest_col = dest_x;

	for ( ; src_col < x2; src_col++, dest_col++) {
	    surface->get_pixel (src_col, src_row, pixel);
	    set_pixel (dest_col, dest_row, pixel);
	}
    }
}


void TessSurface::fade (int percent, int x, int y, int width, int height)
{
    Pixel          pixel;
    int            col, row;
    TessDrawStyle  hold;


    hold = draw_style;  // @@ code a get_draw_style
    set_draw_style (Replace);

    for (row = y; row < y + height; row++) {
	for (col = x; col < x + width; col++) {
	    get_pixel (col, row, pixel);
	    pixel.alpha -= (ccomp) ((float) pixel.alpha * (percent / 100.0));
	    set_pixel (col, row, pixel);
	}
    }

    set_draw_style (hold);
}


void TessSurface::blur (int radius, int x, int y, int width, int height)
{
    Pixel      pixel;
    int        col, row;
    int        red, green, blue, alpha;
    int        blur_col, blur_row;
    int        blur_top, blur_bottom, blur_left, blur_right;
    int        divisor;


#define min(a,b) (a < b ? a : b)
#define max(a,b) (a > b ? a : b)

    assert (radius > 0);

    // @@ "radius" on a rectangle; worth changing to a circle?

    for (row = y; row < y + height; row++) {
	for (col = x; col < x + width; col++) {
	    blur_top    = max (0, row - radius);
	    blur_left   = max (0, col - radius);
	    blur_bottom = min (surface_height - 1, row + radius);
	    blur_right  = min (surface_width  - 1, col + radius);

	    divisor = (blur_bottom - blur_top + 1) *
		(blur_right - blur_left + 1);

	    red = green = blue = alpha = 0;

	    for (blur_row = blur_top; blur_row <= blur_bottom; blur_row++) {
		for (blur_col = blur_left; blur_col <= blur_right; blur_col++){
		    get_pixel (blur_col, blur_row, pixel);
		    
		    red   += pixel.red;
		    green += pixel.green;
		    blue  += pixel.blue;
		    alpha += pixel.alpha;
		}
	    }
	    
	    pixel.red   = red   /= divisor;
	    pixel.green = green /= divisor;
	    pixel.blue  = blue  /= divisor;
	    pixel.alpha = alpha /= divisor;

	    set_pixel (col, row, pixel);
	}
    }
}


void TessSurface::flatten (ccomp red, ccomp green, ccomp blue)
{
    Pixel under;
    int   i;


    tess_setp (under, red, green, blue, 255);

    for (i = 0; i < surface_width * surface_height; i++)
	pixmap [i] = tess_posit (under, pixmap [i]);
}


int TessSurface::write (TessWriter * writer)
{
    writer->set_size (surface_width, surface_height);
    return (writer->commit (this));
}
