//
// tess.hh
//
//  version 0.40  january 17, 1999
//  jeremy wohl (jeremyw@evantide.com)
//  no warranty, no responsibility, etc.
//

#ifndef _tess_hh_
#define _tess_hh_

#include <assert.h>
#include <stdio.h>

#include "freetype.h"

#ifdef SPARC
enum bool { false = 0, true };
#endif

typedef unsigned char ccomp;
struct Pixel {
    ccomp  red, green, blue, alpha;
};

inline void tess_setp (Pixel& pixel, ccomp r, ccomp g, ccomp b, ccomp a)
{ pixel.red = r; pixel.green = g; pixel.blue = b; pixel.alpha = a; }

inline ccomp tess_mult (ccomp a, ccomp b)
{ int t = a * b + 0x80; return (((t >> 8) + t) >> 8); }
inline ccomp tess_pmalerp (ccomp p, ccomp q, ccomp a)
{ return (p + q - tess_mult (a, p)); }
inline Pixel tess_posit (Pixel& under, Pixel &over)
{
    Pixel posit;
    tess_setp (posit, tess_pmalerp (under.red,   over.red,   over.alpha),
	       tess_pmalerp (under.green, over.green, over.alpha),
	       tess_pmalerp (under.blue,  over.blue,  over.alpha),
	       tess_pmalerp (under.alpha, over.alpha, over.alpha));
    return (posit);
}

class TessType;
class TessReader;
class TessWriter;

class TessSurface
{
public:

    enum TessDrawStyle { Blend = 1, Replace       };
    enum TessAlignment { Left  = 1, Right, Center };

private:

    Pixel  * pixmap;
    Pixel    draw_color;
    int      surface_width, surface_height, bezier_intervals;
    int      draw_width, type_inclination;

    TessType       * type;
    TessAlignment    type_alignment;
    TessDrawStyle    draw_style;

    void init (int width, int height);
    
    void draw_naa_line (int x1, int y1, int x2, int y2);
    void draw_aa_line (int x1, int y1, int x2, int y2);

public:

    TessSurface (int width, int height);
    TessSurface (TessReader * reader);

    ~TessSurface ();

    inline void get_size (int& width, int& height)
	{ width = surface_width; height = surface_height; }

    void set_alpha_pixel (int x, int y, ccomp alpha);
    void set_pixel (int x, int y, Pixel& pixel);
    void get_pixel (int x, int y, Pixel& pixel);
    void get_pma_pixel (int x, int y, Pixel& pixel);

    void set_type (TessType * type) { this->type = type; }
    void set_type_alignment (TessAlignment alignment);
    void set_type_inclination (int angle);
    void set_draw_width (int width);
    void set_draw_color (ccomp red, ccomp green, ccomp blue);
    void set_draw_style (TessDrawStyle style);

    void draw_string (int x, int y, char * string);
    void draw_uc_string (int x, int y, short * string);

    void draw_line (int x1, int y1, int x2, int y2);
    void draw_bezier (int x1, int y1, int x2, int y2, int x3, int y3,
		      int x4, int y4);
    void draw_rect (int x, int y, int width, int height);
    void draw_filled_rect (int x, int y, int width, int height);

    void copy_rect (int x, int y, int width, int height,
		    int dest_x, int dest_y);
    void copy_rect (TessSurface * surface, int x, int y, int width, int height,
	       int dest_x, int dest_y);

    void fade (int percent, int x, int y, int width, int height);
    void blur (int radius,  int x, int y, int width, int height);
    void flatten (ccomp red, ccomp green, ccomp blue);

    int write (TessWriter * writer);
};


class TessStack
{
private:

    struct node {
	TessSurface  * surface;
	int            originx, originy;
	int            width, height;
	struct node  * next;
    } * head;

    int maxcol, maxrow;

    inline bool is_layer_at (int col, int row, node * layer);

public:

    TessStack ();
    ~TessStack ();

    void add_surface (TessSurface * surface, int x, int y);
    TessSurface * make_composite ();
};


class TessType
{
private:

    char * newttpath;
    bool   valid, apply_rotation;
    int    no_glyphs, spacing, xcenter, ycenter, ascent, descent, pointsize;

    TT_Face           face;
    TT_Instance       instance;
    TT_Glyph          glyph;
    TT_Raster_Map     map;
    TT_CharMap        charmap;
    TT_Matrix         matrix;

    struct glyph_metric {
	int         advwidth, lsb, rsb, xmin, ymin, xmax, ymax;
	TT_F26Dot6  xoff, yoff;
    } * metrics;

    bool find_font (char * ttpath);
    char * next_token (char ** stringptr);

public:

    TessType (char * ttpath, int pointsize, int dpi_x, int dpi_y);
    ~TessType ();

    bool is_valid () { return (valid); }

    void set_rotation (int angle);
    void set_spacing (int spacing);

    int get_ascent ();
    int get_descent ();
    int get_spacing ();
    int get_string_width (char * string);
    int get_uc_string_width (short * string);

    void render_glyph (TessSurface * surface, int x, int y, int unicode);
    int get_glyph_width (int unicode);
    int get_pair_adjust (int unicode_left, int unicode_right);
};


class TessReader
{
public:

    virtual int get_size (int& width, int& height) = 0;
    virtual int populate (TessSurface * surface)   = 0;

    virtual ~TessReader () { }
};


class TessWriter
{
public:

    virtual void set_size (int width, int height) = 0;
    virtual int commit (TessSurface * surface)    = 0;

    virtual ~TessWriter () { }
};


class TessPPMReader : public TessReader
{
private:

    FILE  * file;
    char  * filepath;
    int     width, height;
    ccomp   alpha;

    int read_int ();

public:

    TessPPMReader (char * filepath, ccomp alpha = 255);
    virtual ~TessPPMReader ();
    
    virtual int get_size (int& width, int& height);
    virtual int populate (TessSurface * surface);
};


class TessPPMWriter : public TessWriter
{
private:

    char  * filepath;
    int     width, height;

public:

    TessPPMWriter (char * filepath);
    virtual ~TessPPMWriter ();

    virtual void set_size (int width, int height);
    virtual int commit (TessSurface * surface);
};


class TessPAMReader : public TessReader
{
private:

    FILE  * file;
    char  * filepath;
    int     width, height;

    int read_int ();

public:

    TessPAMReader (char * filepath);
    virtual ~TessPAMReader ();
    
    virtual int get_size (int& width, int& height);
    virtual int populate (TessSurface * surface);
};


class TessPAMWriter : public TessWriter
{
private:

    char  * filepath;
    int     width, height;

public:

    TessPAMWriter (char * filepath);
    virtual ~TessPAMWriter ();

    virtual void set_size (int width, int height);
    virtual int commit (TessSurface * surface);
};

#endif
