//
// type.cc
//
//  version 0.40  january 17, 1999
//  jeremy wohl (jeremyw@evantide.com)
//  no warranty, no responsibility, etc.
//

#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <math.h>

#include <freetype.h>
#include <ftxkern.h>
#include "tess.hh"
#include "config.hh"

static TT_Engine  engine;
static bool       freetype_initialized = false;
static int        alpha_map [5] = { 0, 63, 127, 191, 255 };


TessType::TessType (char * ttpath, int pointsize, int dpi_x, int dpi_y)
{
    TT_Face_Properties   props;
    TT_Instance_Metrics  imetrics;
    //TT_Kerning           directory;

    bool   charmap_found;
    int    i, index, result;
    int    xcenter_upem, ycenter_upem, units_per_em;
    unsigned short  platform_id, encoding_id;


    spacing         = 0;
    no_glyphs       = 0;
    newttpath       = 0;
    charmap_found   = false;
    valid           = false;
    apply_rotation  = false;

    // @@ reentrancy; call done_freetype on a mutex counter
    if (! freetype_initialized) {
	TT_Init_FreeType (&engine);
	//result = TT_Init_Kerning_Extension (engine);
	//printf ("result: %i\n", result);
	freetype_initialized = true;
    }

    if (find_font (ttpath) == false ||
	TT_Open_Face (engine, newttpath, &face) != 0)  return;

    if (TT_Get_Face_Properties (face, &props))         return;

    result = TT_Get_CharMap_Count (face);

    for (i = 0; i < result; i++) {
	TT_Get_CharMap_ID (face, i, &platform_id, &encoding_id);

	if (platform_id == 3 && encoding_id == 1) {
	    TT_Get_CharMap (face, i, &charmap);
	    charmap_found = true;
	    break;
	}
    }

    if (! charmap_found)  return;
    
#ifdef zed
    result = TT_Get_Kerning_Directory (face, &directory);
    printf ("result: %i\n", result);
    printf ("kern version: %i, nTables: %i\n", directory.version,
	    directory.nTables);
    TT_Load_Kerning_Table (face, kern_index);
#endif

    TT_New_Instance (face, &instance);
    TT_Set_Instance_PointSize (instance, pointsize);
    TT_Set_Instance_Resolutions (instance, dpi_x, dpi_y);
    TT_Get_Instance_Metrics (instance, &imetrics);

    no_glyphs = props.num_Glyphs;

    xcenter_upem  = (props.header->xMax - props.header->xMin) / 2;
    ycenter_upem  = (props.header->yMax - props.header->yMin) / 2;
    units_per_em  = props.header->Units_Per_EM;

    xcenter   = imetrics.x_ppem * xcenter_upem / units_per_em / 4;
    ycenter   = imetrics.y_ppem * ycenter_upem / units_per_em / 4;

    ascent    = (props.horizontal->Ascender  * imetrics.y_ppem) / units_per_em;
    descent   = (props.horizontal->Descender * imetrics.y_ppem) / units_per_em;

    map.rows   = imetrics.y_ppem + 32;
    map.cols   = (imetrics.x_ppem + 32 + 3) & -4;
    map.width  = map.cols;
    map.flow   = TT_Flow_Down;
    map.size   = map.rows * map.cols;
    map.bitmap = (void *) new char [map.size];
    memset (map.bitmap, 0, map.size);

    metrics = new glyph_metric [no_glyphs];

    TT_New_Glyph (face, &glyph);

    for (index = 0; index < no_glyphs; index++) {
	TT_Glyph_Metrics  gmetrics;
	int               advwidth, lsb, xmin, xmax;

	TT_Load_Glyph (instance, glyph, index, TTLOAD_DEFAULT);
	TT_Get_Glyph_Metrics (glyph, &gmetrics);

	metrics [index].advwidth = advwidth = gmetrics.advance  >> 6;
	metrics [index].lsb      = lsb      = gmetrics.bearingX >> 6;
	metrics [index].xmin     = xmin     = gmetrics.bbox.xMin >> 6;
	metrics [index].ymin                = gmetrics.bbox.yMin >> 6;
	metrics [index].xmax     = xmax     = gmetrics.bbox.xMax >> 6;
	metrics [index].ymax                = gmetrics.bbox.yMax >> 6;
	metrics [index].rsb      = advwidth - (lsb + (xmax - xmin));

	metrics [index].xoff = -(gmetrics.bbox.xMin & -64);
	metrics [index].yoff = -(gmetrics.bbox.yMin & -64);
    }

    valid = true;

    return;
}


bool TessType::find_font (char * ttpath)
{
    struct stat    statbuf;
    char         * p, * trypath, * token;
    bool           first_check = true, done = false;


    if (stat (ttpath, &statbuf) == 0) {
	newttpath = new char [strlen (ttpath) + 1];
	strcpy (newttpath, ttpath);
	return (true);
    }

    //
    // try default font path, then try TTFONTS environment variable
    //
    while (! done) {
	if (first_check) {
	    p = (char *) DefaultFontPath;
	    first_check = false;
	}
	else {
	    if ((p = getenv ("TTFONTS")) == 0)  return false;
	    done = true;
	}

	while ((token = next_token (&p))) {
	    trypath = new char [strlen (token) + 1 + strlen (ttpath) + 1];
	    
	    strcpy (trypath, token);
	    strcat (trypath, "/");
	    strcat (trypath, ttpath);

	    delete token;

	    if (stat (trypath, &statbuf) == 0) {
		newttpath = trypath;
		return (true);
	    }

	    delete trypath;
	}
    }

    return (false);
}


char * TessType::next_token (char ** stringptr)
{
    int     len;
    char  * start, * end, * token;


    if (*stringptr [0] == 0)
	return 0;

    start = *stringptr;

    for (end = *stringptr; *end != ':' && *end != 0 ; end++)
	;

    len = end - start;

    token = new char [len + 1];
    strncpy (token, *stringptr, len);
    token [len] = 0;

    if (*end == 0)
	*stringptr = end;
    else
	*stringptr = end + 1;

    return token;
}


TessType::~TessType ()
{
    if (map.bitmap)  delete map.bitmap;
    if (newttpath)   delete newttpath;

    TT_Close_Face (face);
}


int TessType::get_ascent ()
{
    return (ascent);
}


int TessType::get_descent ()
{
    return (descent);
}


void TessType::set_rotation (int angle)
{
    double radians;


    if (angle == 0) {
	TT_Set_Instance_Transform_Flags (instance, 0, 0);
	
	apply_rotation = false;
    }
    else {
	// @@ avoid, since hinting is turned off in many cases?
	TT_Set_Instance_Transform_Flags (instance, 1, 0);

	angle  %= 360;
	radians = angle * M_PI / 180.0;

	matrix.xx = (TT_Fixed) (cos (radians) * (1 << 16));
	matrix.xy = (TT_Fixed) (sin (radians) * (1 << 16));
	matrix.yx = - matrix.xy;
	matrix.yy = matrix.xx;

	apply_rotation = true;
    }
}


void TessType::set_spacing (int spacing)
{
    this->spacing = spacing;
}


int TessType::get_spacing ()
{
    return (spacing);
}


int TessType::get_string_width (char * string)
{
    int width = 0, no_chars = 0, uc;

    // @@ kerning
    while ((uc = *string++))  { width += get_glyph_width (uc); no_chars++; }
    width += (no_chars - 1) * spacing;
    width -= metrics [TT_Char_Index (charmap, *(string - 2))].rsb;

    return (width);
}


int TessType::get_uc_string_width (short * string)
{
    int width = 0, no_chars = 0, uc;

    while ((uc = *string++))  { width += get_glyph_width (uc); no_chars++; }
    width += (no_chars - 1) * spacing;
    width -= metrics [TT_Char_Index (charmap, *(string - 2))].rsb;

    return (width);
}


void TessType::render_glyph (TessSurface * surface, int x, int y, int unicode)
{
    //TT_Glyph_Outline   outline;

    char  * bitmap;
    int     col, row, glyph_index, result, cell;


    glyph_index = TT_Char_Index (charmap, unicode);
    result = TT_Load_Glyph (instance, glyph, glyph_index, TTLOAD_DEFAULT);

    if (apply_rotation) {
	//TT_Get_Glyph_Outline (glyph, &outline);
	//TT_Translate_Outline (&outline, -xcenter * 64, -ycenter * 64);
	//TT_Transform_Outline (&outline, &matrix);
	//TT_Translate_Outline (&outline, xcenter * 64, ycenter * 64);
    }

    y -= map.rows;

    memset (map.bitmap, 0, map.size);

    TT_Get_Glyph_Pixmap (glyph, &map,
			 metrics [glyph_index].xoff,
			 metrics [glyph_index].yoff);

    bitmap = (char *) map.bitmap;

    for (row = 0; row < map.rows; row++) {
	for (col = 0; col < map.cols; col++) {
	    cell = bitmap [row * map.cols + col];

	    if (cell)
		surface->set_alpha_pixel
		    (x + col +  metrics [glyph_index].xmin,
		     y + row + -metrics [glyph_index].ymin,
		     alpha_map [cell]);
	}
    }
}


int TessType::get_glyph_width (int unicode)
{
    return (metrics [TT_Char_Index (charmap, unicode)].advwidth);
}


int TessType::get_pair_adjust (int unicode_left, int unicode_right)
{
    return 0;
}
